/*
 * file XmlTagVisitor.java
 *
 * IBM Confidential
 * OCO Source Materials
 * Rational ClearCase
 * (C) Copyright IBM Corp. 2010.  All rights reserved.
 *
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
 */
package com.ibm.rational.wvcm.stpex.annotation.processing;

import static com.ibm.rational.wvcm.stpex.annotation.processing.ListClassesVisitor.INDENT;
import static com.ibm.rational.wvcm.stpex.annotation.processing.ListClassesVisitor.createFile;
import static com.ibm.rational.wvcm.stpex.annotation.processing.ListClassesVisitor.ensureUniqueAndAddIfSo;
import static com.ibm.rational.wvcm.stpex.annotation.processing.ListClassesVisitor.getRpcEnumValues;
import static com.ibm.rational.wvcm.stpex.annotation.processing.ListClassesVisitor.writeCommonHeader;

import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeSet;

import com.sun.mirror.apt.AnnotationProcessorEnvironment;
import com.sun.mirror.apt.Filer;
import com.sun.mirror.declaration.FieldDeclaration;
import com.sun.mirror.util.SimpleDeclarationVisitor;

public class XmlTagVisitor extends SimpleDeclarationVisitor implements StpVisitor
{
    private final AnnotationProcessorEnvironment m_env;
    
    private static final String g_hxxFileName = "xmltag_tag_decl.hxx";
    private static final String g_cxxFileName = "xmltag_tag_impl.cxx";
    private static final String g_xFileName = "stprpc_xml_tag_id.x";
    private static final String g_javaFileName = "XmlTagId.java";
    
    private PrintWriter m_hxxOut;
    private PrintWriter m_cxxOut;
    private PrintWriter m_xOut;

    private PrintWriter m_javaOut;
    
    private final Map<Integer, TagInfo> m_xmltags = new HashMap<Integer, TagInfo>();
    
    public XmlTagVisitor(AnnotationProcessorEnvironment env)
    {
        //System.out.println("Creating XmlTagVistor");
        m_env = env;
    }
    
    public void begin()
    {
        //System.out.println("XmlTagVistor begin");
        Filer filer = m_env.getFiler();

        writeHxxBegin(filer);
        writeCxxBegin(filer);
        writeXBegin(filer);

        writeJavaBegin(filer);
    }        
    
    public void end()
    {
        TreeSet<Integer> keys = new TreeSet<Integer>(m_xmltags.keySet());

        writeHxxEnd(keys);
        writeCxxEnd(keys);
        writeXEnd(keys);

        writeJavaEnd(keys);

        //System.out.println("XmlTagVistor end");
    }
    
    public void writeXBegin(Filer filer) 
    {
        m_xOut = createFile(filer, g_xFileName);
        writeCommonHeader(m_xOut, this);
    }
    
    public void writeXEnd(TreeSet<Integer> keys)
    {
        for (Integer id : keys) {
            TagInfo info = m_xmltags.get(id);
            m_xOut.println(INDENT + info.fieldName + " = " + id + ",");
        }
        
        m_xOut.println(INDENT + "FIRST_DYNAMIC_PROP = 1000000");
        
        m_xOut.flush();
        m_xOut.close();      
    }
    
    public void writeHxxBegin(Filer filer)
    {
        m_hxxOut = createFile(filer, g_hxxFileName);
        writeCommonHeader(m_hxxOut, this);
        m_hxxOut.println("");
        m_hxxOut.println("enum Id {");
    }
    
    public void writeHxxEnd(TreeSet<Integer> keys)
    {
        for (Integer id : keys) {
            TagInfo info = m_xmltags.get(id);
            m_hxxOut.println(INDENT + info.fieldName + " = " + id + ",");
        }
        
        m_hxxOut.println(INDENT + "NUM_PREDEFINED_TAGS");
        m_hxxOut.println("};");
        
        m_hxxOut.flush();
        m_hxxOut.close();        
    }
    
    public void writeCxxBegin(Filer filer) 
    {
        m_cxxOut = createFile(filer, g_cxxFileName);
        
        writeCommonHeader(m_cxxOut, this);
    }
    
    public void writeCxxEnd(TreeSet<Integer> keys)
    {
        for (Integer id : keys) {
            TagInfo info = m_xmltags.get(id);
            m_cxxOut.println("TAG_DEFN(L\"" + info.ns + "\", L\"" + info.simple + "\", " + info.fieldName + ");"); 
        }
        
        m_cxxOut.flush();
        m_cxxOut.close();        
    }

    public void writeJavaBegin(Filer filer)
    {
        m_javaOut = createFile(filer, g_javaFileName);

        writeCommonHeader(m_javaOut, this);

        m_javaOut.println("package com.ibm.rational.stp.server.rpc;");
        m_javaOut.println("");
        m_javaOut.println("import java.util.Map;");
        m_javaOut.println("import java.util.Map.Entry;");
        m_javaOut.println("import java.util.HashMap;");
        m_javaOut.println("import com.ibm.rational.stp.cs.internal.util.XmlTag;");
        m_javaOut.println("/**");
        m_javaOut.println(" * Mapping services XmlTag objects and RPC XML tag ID codes.");
        m_javaOut.println(" */");
        m_javaOut.println("public class XmlTagId");
        m_javaOut.println("{");
        m_javaOut.println("    public static final int FIRST_DYNAMIC_PROP = 1000000;");
        m_javaOut.println("    /**");
        m_javaOut.println("     * Get the XML tag associated with integer tag ID.");
        m_javaOut.println("     *");
        m_javaOut.println("     * @param name IN: RPC integer value tag ID");
        m_javaOut.println("     *");
        m_javaOut.println("     * @return XmlTag associated with RPC tag ID");
        m_javaOut.println("     */");
        m_javaOut.println("    public static XmlTag");
        m_javaOut.println("    getTag(int id) ");
        m_javaOut.println("    {");
        m_javaOut.println("        XmlTag tag = g_idToTagMap.get(id);");
        m_javaOut.println("");
        m_javaOut.println("        return tag == null ? XmlTag.NIL : tag;");
        m_javaOut.println("    }");
        m_javaOut.println("");
        m_javaOut.println("    /**");
        m_javaOut.println("     * Get RPC integer value tag ID associated with XML tag.");
        m_javaOut.println("     *");
        m_javaOut.println("     * @param name IN: XmlTag");
        m_javaOut.println("     *");
        m_javaOut.println("     * @return RPC integer value tag ID");
        m_javaOut.println("     */");
        m_javaOut.println("    public static int");
        m_javaOut.println("    getId(XmlTag tag)");
        m_javaOut.println("    {");
        m_javaOut.println("        Integer id = g_tagToIdMap.get(tag);");
        m_javaOut.println("");
        m_javaOut.println("        return id == null ? 0 : id.intValue();");
        m_javaOut.println("    }");
        m_javaOut.println("");
        m_javaOut.println("    private static final Map<Integer, XmlTag> g_idToTagMap = new HashMap<Integer, XmlTag>();");
        m_javaOut.println("    private static final Map<XmlTag, Integer> g_tagToIdMap = new HashMap<XmlTag, Integer>();");
        m_javaOut.println("");
        m_javaOut.println("    static {");
    }

    public void writeJavaEnd(TreeSet<Integer> keys)
    {
        for (Integer id : keys) {
            TagInfo info = m_xmltags.get(id);
            m_javaOut.println("        g_idToTagMap.put(" + info.id + ", XmlTag." + info.fieldName + ");");
        }

        m_javaOut.println("");
        m_javaOut.println("");
        m_javaOut.println("        for (Entry<Integer, XmlTag> entry: g_idToTagMap.entrySet()) {");
        m_javaOut.println("            g_tagToIdMap.put(entry.getValue(), entry.getKey());");
        m_javaOut.println("        }");
        m_javaOut.println("    }");
        m_javaOut.println("}");

        m_javaOut.flush();
        m_javaOut.close();
    }
    
    @Override
    public void visitFieldDeclaration(FieldDeclaration f) 
    {
        TagInfo info = getRpcEnumValues(f.getAnnotationMirrors());
        
        if (info != null) {
        
            info.fieldName = f.getSimpleName();
            
            if (info.id == null || info.ns == null || info.simple == null) {
                
                System.err.print("Error: missing value for field '"+ info.fieldName + "' no values should be null: ");
                System.err.println("id = " + info.id + ", ns = " + info.ns + ", simple = " + info.simple);
                System.err.println("Error: exiting unexpectedly");
                System.exit(0);
            }
            
            ensureUniqueAndAddIfSo(m_xmltags, info);
        }
    }
}
